﻿Option Strict On
Option Explicit On

Imports System.Runtime.InteropServices

Public Class Mp3OutputProfiles

#Region "Implements"
    Implements IDisposable
#End Region

#Region "APIs"

    ' ----==== Mfplat.dll ====----
    <DllImport("Mfplat.dll", EntryPoint:="MFStartup")>
    <PreserveSig> Private Shared Function MFStartup(<[In]> Version As Integer,
                                                    <[In]> dwFlags As Integer) As Integer
    End Function

    <DllImport("Mfplat.dll", EntryPoint:="MFShutdown")>
    <PreserveSig> Private Shared Function MFShutdown() As Integer
    End Function

    <DllImport("Mfplat.dll", EntryPoint:="MFTEnumEx")>
    <PreserveSig> Private Shared Function MFTEnumEx(<[In]> guidCategory As Guid,
                                                    <[In]> flags As MFT_ENUM_FLAG,
                                                    <[In]> pInputType As IntPtr,
                                                    <[In]> pOutputType As IntPtr,
                                                    <Out> ByRef pppMFTActivate As IntPtr,
                                                    <Out> ByRef pnumMFTActivate As Integer) As Integer
    End Function

    ' ----==== Ole32.dll ====----
    <DllImport("Ole32.dll", EntryPoint:="PropVariantClear")>
    <PreserveSig> Private Shared Function PropVariantClear(<[In], Out> ByRef pvar As PROPVARIANT) As Integer
    End Function

#End Region

#Region "Class"
    Public Sub New()

        ' die Media Foundation ist noch nicht initialisiert
        If Not m_MFInit Then

            ' Media Foundation initialisiern
            If MFStartup(MF_VERSION_2, MFSTARTUP_FULL) = S_OK Then

                m_MFInit = True

            End If

        End If

    End Sub
#End Region

#Region "Const"
    Private Const S_OK As Integer = &H0

    Private Const MFSTARTUP_FULL As Integer = &H0
    Private Const MF_VERSION_2 As Integer = &H20070

    Private Const MF_MT_SUBTYPE As String = "f7e34c9a-42e8-4714-b74b-cb29d72c35e5"
    Private Const MF_MT_MAJOR_TYPE As String = "48eba18e-f8c9-4687-bf11-0a74c9f96a8f"

    Private Const MF_MT_AUDIO_NUM_CHANNELS As String = "37e48bf5-645e-4c5b-89de-ada9e29b696a"
    Private Const MF_MT_AUDIO_BLOCK_ALIGNMENT As String = "322de230-9eeb-43bd-ab7a-ff412251541d"
    Private Const MF_MT_AUDIO_SAMPLES_PER_SECOND As String = "5faeeae7-0290-4c31-9e8a-c534f68d9dba"
    Private Const MF_MT_AUDIO_AVG_BYTES_PER_SECOND As String = "1aab75c8-cfef-451c-ab95-ac034b8e1731"

    Private Const MFMediaType_Audio As String = "73647561-0000-0010-8000-00aa00389b71"
    Private Const MFAudioFormat_MP3 As String = "00000055-0000-0010-8000-00aa00389b71"

    Private Const MFT_CATEGORY_AUDIO_ENCODER As String = "91c64bd0-f91e-4d8c-9276-db248279d975"

    Private Const IID_IMFTransform As String = "bf94c121-5b05-4e6f-8000-ba598961414d"
#End Region

#Region "Enums"
    Private Enum VTable_Interfaces As Integer

        ' IMFActivate
        ActivateObject = 33

        ' IMFTransform
        GetOutputAvailableType = 14

        ' IMFMediaType / IMFAttributes
        GetGUID = 10
        GetCount = 30
        GetItemByIndex = 31

    End Enum

    Private Enum MFT_ENUM_FLAG As Integer
        MFT_ENUM_FLAG_SYNCMFT = &H1
        MFT_ENUM_FLAG_ASYNCMFT = &H2
        MFT_ENUM_FLAG_HARDWARE = &H4
        MFT_ENUM_FLAG_FIELDOFUSE = &H8
        MFT_ENUM_FLAG_LOCALMFT = &H10
        MFT_ENUM_FLAG_TRANSCODE_ONLY = &H20
        MFT_ENUM_FLAG_SORTANDFILTER = &H40
        MFT_ENUM_FLAG_SORTANDFILTER_APPROVED_ONLY = &HC0
        MFT_ENUM_FLAG_SORTANDFILTER_WEB_ONLY = &H140
        MFT_ENUM_FLAG_SORTANDFILTER_WEB_ONLY_EDGEMODE = &H240
        MFT_ENUM_FLAG_UNTRUSTED_STOREMFT = &H400
        MFT_ENUM_FLAG_ALL = &H3F
    End Enum

    Private Enum MF_ATTRIBUTE_TYPE As UShort
        MF_ATTRIBUTE_DOUBLE = 5
        MF_ATTRIBUTE_IUNKNOWN = 13
        MF_ATTRIBUTE_UINT32 = 19
        MF_ATTRIBUTE_UINT64 = 21
        MF_ATTRIBUTE_STRING = 31
        MF_ATTRIBUTE_GUID = 72
        MF_ATTRIBUTE_BLOB = 4113
    End Enum
#End Region

#Region "Structure"
    Private Structure CArray
        Dim cElems As UInteger
        Dim pElems As IntPtr
    End Structure

    <StructLayout(LayoutKind.Explicit, Size:=16)>
    Private Structure PROPVARIANT
        <FieldOffset(0)> Dim vt As MF_ATTRIBUTE_TYPE
        <FieldOffset(2)> Dim wReserved1 As UShort
        <FieldOffset(4)> Dim wReserved2 As UShort
        <FieldOffset(6)> Dim wReserved3 As UShort
        <FieldOffset(8)> Dim uintVal As UInteger 'UINT32
        <FieldOffset(8)> Dim ulVal As ULong      'UINT64
        <FieldOffset(8)> Dim dblVal As Double    'DOUBLE
        <FieldOffset(8)> Dim pwszVal As IntPtr   'STRING
        <FieldOffset(8)> Dim punkVal As IntPtr   'IUNKNOWN
        <FieldOffset(8)> Dim cArray As CArray    'BLOB
    End Structure

    Public Structure MP3_OUTPUT_PROFILE
        Dim MediaSubType As Guid
        Dim MediaMajorType As Guid
        Dim AudioNumChannels As UInt32
        Dim AudioBlockAlignment As UInt32
        Dim AudioSamplesPerSecond As UInt32
        Dim AudioAvgBytesPerSecond As UInt32
    End Structure
#End Region

#Region "Variablen"
    Private m_MFInit As Boolean = False
    Private m_DisposedValue As Boolean = False
    Private m_Mp3OutputProfiles As List(Of MP3_OUTPUT_PROFILE)
#End Region

#Region "Properties"
    Public ReadOnly Property IsInitialized As Boolean
        Get
            Return m_MFInit
        End Get
    End Property

    Public ReadOnly Property GetMp3ProfileByIndex(Index As Integer) As MP3_OUTPUT_PROFILE
        Get
            Return m_Mp3OutputProfiles(Index)
        End Get
    End Property

#End Region

#Region "Public Functions"

    Public Function ListMp3Profiles() As Integer

        Dim intActivateCount As Integer
        Dim pIMFActivateArr As IntPtr = IntPtr.Zero
        Dim eFlags As MFT_ENUM_FLAG = MFT_ENUM_FLAG.MFT_ENUM_FLAG_ALL And Not MFT_ENUM_FLAG.MFT_ENUM_FLAG_FIELDOFUSE

        ' neue Liste 
        m_Mp3OutputProfiles = New List(Of MP3_OUTPUT_PROFILE)

        ' alle Audio-Encoder auflisten -> Pointer auf ein IMFActivate-Array
        If MFTEnumEx(New Guid(MFT_CATEGORY_AUDIO_ENCODER), eFlags, IntPtr.Zero,
                     IntPtr.Zero, pIMFActivateArr, intActivateCount) = S_OK Then

            ' sind Audio-Encoder vorhanden
            If intActivateCount > 0 Then

                ' IntPtr-Array, zur Aufnahme der IMFActivate-Pointer, dimensionieren
                Dim pIMFActivate As IntPtr() = New IntPtr(intActivateCount - 1) {}

                ' IMFActivate-Pointer aus dem IMFActivate-Array umkopieren
                Marshal.Copy(pIMFActivateArr, pIMFActivate, 0, intActivateCount)

                ' alle IMFActivate durchlaufen
                For intActivateItem As Integer = 0 To intActivateCount - 1

                    Dim pIMFTransform As IntPtr = IntPtr.Zero

                    ' vom IMFActivate einen IMFTransform erstellen
                    If MFActivate_ActivateObject(pIMFActivate(intActivateItem),
                                                 New Guid(IID_IMFTransform), pIMFTransform) Then

                        Dim intTypeIndex As Integer = 0
                        Dim pIMFMediaType As IntPtr = IntPtr.Zero

                        ' alle Outputs vom IMFTransform durchlaufen und ein IMFMediaType erstellen
                        Do While MFTransform_GetOutputAvailableType(pIMFTransform, 0, intTypeIndex, pIMFMediaType)

                            Dim tGuid As New Guid

                            ' MajorType vom IMFMediaType ermitteln
                            If MFMediaType_GetGUID(pIMFMediaType, New Guid(MF_MT_MAJOR_TYPE), tGuid) Then

                                ' ist der MajorType für Audio
                                If tGuid = New Guid(MFMediaType_Audio) Then

                                    ' SubType vom IMFMediaType ermitteln
                                    If MFMediaType_GetGUID(pIMFMediaType, New Guid(MF_MT_SUBTYPE), tGuid) Then

                                        ' ist der SubType für MP3
                                        If tGuid = New Guid(MFAudioFormat_MP3) Then

                                            Dim intMediaCount As Integer

                                            ' Anzahl der Eigenschaften für diesen IMFMediaType ermitteln
                                            If MFMediaType_GetCount(pIMFMediaType, intMediaCount) Then

                                                ' sind Eigenschaften vorhanden
                                                If intMediaCount > 0 Then

                                                    ' neues Output Profile
                                                    Dim OutputProfile As New MP3_OUTPUT_PROFILE

                                                    ' alle Eigenschaften durchlaufen
                                                    For intMediaItem = 0 To intMediaCount - 1

                                                        Dim tAttributeGuid As New Guid
                                                        Dim tPROPVARIANT As New PROPVARIANT

                                                        ' Eigenschaft über den Index ermitteln
                                                        If MFMediaType_GetItemByIndex(pIMFMediaType, intMediaItem,
                                                                                      tAttributeGuid, tPROPVARIANT) Then

                                                            ' Stark eingekürzt und speziell für MP3! Daher weis ich auch welcher
                                                            ' tPROPVARIANT.vt für was steht bzw. wie dieser zu konvertieren ist!
                                                            ' Es gibt noch die Eigenschaften MF_MT_AUDIO_PREFER_WAVEFORMATEX und
                                                            ' MF_MT_USER_DATA für MP3. Diese benötigt man aber nicht unbedingt.

                                                            ' nach der GUID der Eigenschaft selektieren
                                                            Select Case tAttributeGuid
                                                                Case New Guid(MF_MT_MAJOR_TYPE)
                                                                    If tPROPVARIANT.vt = MF_ATTRIBUTE_TYPE.MF_ATTRIBUTE_GUID Then
                                                                        OutputProfile.MediaMajorType = New Guid(MFMediaType_Audio)
                                                                    End If

                                                                Case New Guid(MF_MT_SUBTYPE)
                                                                    If tPROPVARIANT.vt = MF_ATTRIBUTE_TYPE.MF_ATTRIBUTE_GUID Then
                                                                        OutputProfile.MediaSubType = New Guid(MFAudioFormat_MP3)
                                                                    End If

                                                                Case New Guid(MF_MT_AUDIO_AVG_BYTES_PER_SECOND)
                                                                    If tPROPVARIANT.vt = MF_ATTRIBUTE_TYPE.MF_ATTRIBUTE_UINT32 Then
                                                                        OutputProfile.AudioAvgBytesPerSecond = tPROPVARIANT.uintVal
                                                                    End If

                                                                Case New Guid(MF_MT_AUDIO_BLOCK_ALIGNMENT)
                                                                    If tPROPVARIANT.vt = MF_ATTRIBUTE_TYPE.MF_ATTRIBUTE_UINT32 Then
                                                                        OutputProfile.AudioBlockAlignment = tPROPVARIANT.uintVal
                                                                    End If

                                                                Case New Guid(MF_MT_AUDIO_NUM_CHANNELS)
                                                                    If tPROPVARIANT.vt = MF_ATTRIBUTE_TYPE.MF_ATTRIBUTE_UINT32 Then
                                                                        OutputProfile.AudioNumChannels = tPROPVARIANT.uintVal
                                                                    End If

                                                                Case New Guid(MF_MT_AUDIO_SAMPLES_PER_SECOND)
                                                                    If tPROPVARIANT.vt = MF_ATTRIBUTE_TYPE.MF_ATTRIBUTE_UINT32 Then
                                                                        OutputProfile.AudioSamplesPerSecond = tPROPVARIANT.uintVal
                                                                    End If

                                                            End Select

                                                            ' PROPVARIANT löschen
                                                            PropVariantClear(tPROPVARIANT)

                                                        End If

                                                    Next

                                                    ' Output Profile der Liste hinzufügen
                                                    m_Mp3OutputProfiles.Add(OutputProfile)

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                            ' IMFMediaType löschen
                            Marshal.Release(pIMFMediaType)

                            ' nächster IMFMediaType
                            intTypeIndex += 1

                        Loop

                        ' IMFTransform löschen
                        Marshal.Release(pIMFTransform)

                    End If

                    ' IMFActivate löschen
                    Marshal.Release(pIMFActivate(intActivateItem))

                Next

            End If

            ' Speicher auf das IMFActivate-Array freigeben
            Marshal.FreeCoTaskMem(pIMFActivateArr)

        End If

        ' sind Einträge in der Liste vorhanden
        If m_Mp3OutputProfiles.Count > 0 Then

            ' Liste sortieren
            m_Mp3OutputProfiles = m_Mp3OutputProfiles.OrderByDescending(Function(x) x.AudioAvgBytesPerSecond).ThenByDescending(Function(x) x.AudioSamplesPerSecond).ToList()

        End If

        ' Anzahl der Elemente in der Liste zurück geben
        Return m_Mp3OutputProfiles.Count

    End Function
#End Region

#Region "Interface Delegates"
    Private Delegate Function IMFActivate_ActivateObject(<[In]> this As IntPtr,
                                                         <[In]> ByRef riid As Guid,
                                                         <Out> ByRef ppv As IntPtr) As Integer

    Private Delegate Function IMFTransform_GetOutputAvailableType(<[In]> this As IntPtr,
                                                                  <[In]> dwOutputStreamID As Integer,
                                                                  <[In]> dwTypeIndex As Integer,
                                                                  <Out> ByRef ppType As IntPtr) As Integer

    Private Delegate Function IMFMediaType_GetGUID(<[In]> this As IntPtr,
                                                   <[In]> ByRef guidKey As Guid,
                                                   <Out> ByRef pguidValue As Guid) As Integer

    Private Delegate Function IMFMediaType_GetCount(<[In]> this As IntPtr,
                                                    <Out> ByRef pcItems As Integer) As Integer

    Private Delegate Function IMFMediaType_GetItemByIndex(<[In]> this As IntPtr,
                                                          <[In]> unIndex As Integer,
                                                          <Out> ByRef pguidKey As Guid,
                                                          <Out> ByRef pValue As PROPVARIANT) As Integer
#End Region

#Region "Interfcae Functions"
    Private Function MFActivate_ActivateObject(this As IntPtr,
                                               riid As Guid,
                                               ByRef ppv As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.ActivateObject)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFActivate_ActivateObject = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                      GetType(IMFActivate_ActivateObject)),
                                                                      IMFActivate_ActivateObject)

                If Method(this, riid, ppv) = S_OK Then Return True

            End If

        End If

        Return False

    End Function

    Private Function MFTransform_GetOutputAvailableType(this As IntPtr,
                                                        dwOutputStreamID As Integer,
                                                        dwTypeIndex As Integer,
                                                        ByRef ppType As IntPtr) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.GetOutputAvailableType)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFTransform_GetOutputAvailableType = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                               GetType(IMFTransform_GetOutputAvailableType)),
                                                                               IMFTransform_GetOutputAvailableType)

                If Method(this, dwOutputStreamID, dwTypeIndex, ppType) = S_OK Then Return True

            End If

        End If

        Return False

    End Function

    Private Function MFMediaType_GetGUID(this As IntPtr,
                                         guidKey As Guid,
                                         ByRef pguidValue As Guid) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.GetGUID)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFMediaType_GetGUID = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                GetType(IMFMediaType_GetGUID)),
                                                                IMFMediaType_GetGUID)

                If Method(this, guidKey, pguidValue) = S_OK Then Return True

            End If

        End If

        Return False

    End Function

    Private Function MFMediaType_GetCount(this As IntPtr,
                                          ByRef pcItems As Integer) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.GetCount)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFMediaType_GetCount = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                 GetType(IMFMediaType_GetCount)),
                                                                 IMFMediaType_GetCount)

                If Method(this, pcItems) = S_OK Then Return True

            End If

        End If

        Return False

    End Function

    Private Function MFMediaType_GetItemByIndex(this As IntPtr,
                                                unIndex As Integer,
                                                ByRef pguidKey As Guid,
                                                ByRef pValue As PROPVARIANT) As Boolean

        If this <> IntPtr.Zero Then

            Dim pMethodPtr As IntPtr = GetMethodPtr(this, VTable_Interfaces.GetItemByIndex)

            If pMethodPtr <> IntPtr.Zero Then

                Dim Method As IMFMediaType_GetItemByIndex = DirectCast(Marshal.GetDelegateForFunctionPointer(pMethodPtr,
                                                                       GetType(IMFMediaType_GetItemByIndex)),
                                                                       IMFMediaType_GetItemByIndex)

                If Method(this, unIndex, pguidKey, pValue) = S_OK Then Return True

            End If

        End If

        Return False

    End Function


    Private Function GetMethodPtr(pInterface As IntPtr,
                                  intMethodNumber As Integer) As IntPtr

        Dim pMethodPtr As IntPtr = IntPtr.Zero

        If pInterface <> IntPtr.Zero And intMethodNumber >= 0 Then

            Dim pMethod As IntPtr() = New IntPtr(0) {}

            Marshal.Copy(IntPtr.Add(Marshal.ReadIntPtr(pInterface),
                         IntPtr.Size * intMethodNumber), pMethod, 0, 1)

            pMethodPtr = pMethod(0)

        End If

        Return pMethodPtr

    End Function
#End Region

#Region "Implements IDisposable"
    Protected Overridable Sub Dispose(disposing As Boolean)

        If Not m_DisposedValue Then

            If disposing Then

                ' die Media Foundation ist initialisiert
                If m_MFInit Then

                    ' Media Foundation ist noch nicht initialisiert
                    If MFShutdown() = S_OK Then m_MFInit = False

                End If

            End If

            m_DisposedValue = True

        End If

    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose

        Dispose(disposing:=True)

        GC.SuppressFinalize(Me)

    End Sub
#End Region

End Class
